==> gzochi-0.3.tar.gz: gzochi-0.3/gzochi-server/src/schedule.c <==
/* schedule.c: Task execution and task queue management routines for gzochid
 * Copyright (C) 2013 Julian Graham
 *
 * gzochi is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <glib.h>

#include "guile.h"
#include "schedule.h"
#include "task.h"

static void pending_task_executor (gpointer data, gpointer user_data)
{
  gzochid_pending_task *pending_task = (gzochid_pending_task *) data;

  g_mutex_lock (pending_task->mutex);
  pending_task->task->worker (pending_task->task->data, user_data);
  pending_task->state = GZOCHID_PENDING_TASK_STATE_COMPLETED;
  g_cond_broadcast (pending_task->cond);
  g_mutex_unlock (pending_task->mutex);
}

gpointer gzochid_schedule_task_executor (gpointer data)
{
  gzochid_task_queue *task_queue = (gzochid_task_queue *) data;

  g_mutex_lock (task_queue->mutex);
  while (TRUE)
    {
      if (g_queue_is_empty (task_queue->queue))
	g_cond_wait (task_queue->cond, task_queue->mutex);
      else 
	{
	  struct timeval current_time;
	  gzochid_pending_task *pending_task = 
	    (gzochid_pending_task *) g_queue_peek_head (task_queue->queue);

	  gettimeofday (&current_time, NULL);

	  if (timercmp 
	      (&current_time, &pending_task->scheduled_execution_time, >))
	    {
	      pending_task = (gzochid_pending_task *) 
		g_queue_pop_head (task_queue->queue);
	      gzochid_guile_thread_pool_push 
		(task_queue->pool, pending_task_executor, pending_task, NULL);
	    }
	  else 
	    {
	      struct timeval interval;
	      gint64 until = g_get_monotonic_time ();
	      
	      timersub 
		(&pending_task->scheduled_execution_time, 
		 &current_time, 
		 &interval);

	      until += interval.tv_sec * G_TIME_SPAN_SECOND + interval.tv_usec;
	      g_cond_wait_until (task_queue->cond, task_queue->mutex, until);
	    }
	}
    }
  g_mutex_unlock (task_queue->mutex);

  return NULL;
}

void gzochid_schedule_task_queue_start (gzochid_task_queue *task_queue)
{
  if (task_queue->consumer_thread == NULL)
    task_queue->consumer_thread = g_thread_create 
      (gzochid_schedule_task_executor, task_queue, FALSE, NULL);
}

gzochid_pending_task *gzochid_pending_task_new (gzochid_task *task)
{
  gzochid_pending_task *pending_task = malloc (sizeof (gzochid_pending_task));

  pending_task->task = task;
  pending_task->state = GZOCHID_PENDING_TASK_STATE_PENDING;
  pending_task->cond = g_cond_new ();
  pending_task->mutex = g_mutex_new ();

  pending_task->scheduled_execution_time = task->target_execution_time;
  
  return pending_task;
}

gint pending_task_compare (gconstpointer a, gconstpointer b, gpointer user_data)
{
  gzochid_pending_task *pending_task_a = (gzochid_pending_task *) a;
  gzochid_pending_task *pending_task_b = (gzochid_pending_task *) b;

  if (pending_task_a->scheduled_execution_time.tv_sec <
      pending_task_b->scheduled_execution_time.tv_sec)
    return -1;
  else if (pending_task_a->scheduled_execution_time.tv_sec 
	   > pending_task_b->scheduled_execution_time.tv_sec)
    return 1;
  else return pending_task_a->scheduled_execution_time.tv_usec 
	 - pending_task_b->scheduled_execution_time.tv_usec;
}

static void task_chain_worker (gpointer data, gpointer user_data)
{
  GList *tasks = (GList *) data;
  GList *task_ptr = tasks;

  while (task_ptr != NULL)
    {
      gzochid_schedule_execute_task ((gzochid_task *) task_ptr->data);
      task_ptr = task_ptr->next;
    }

  g_list_free (tasks);
}

gzochid_pending_task *gzochid_schedule_submit_task_chain
(gzochid_task_queue *task_queue, GList *tasks)
{
  GList *tasks_copy = g_list_copy (tasks);
  gzochid_task *task = malloc (sizeof (gzochid_task));

  assert (g_list_length (tasks) > 0);

  task->worker = task_chain_worker;
  task->data = tasks_copy;
  task->target_execution_time = 
    ((gzochid_task *) tasks_copy->data)->target_execution_time;

  return gzochid_schedule_submit_task (task_queue, task);
}

gzochid_pending_task *gzochid_schedule_submit_task
(gzochid_task_queue *task_queue, gzochid_task *task)
{
  gzochid_pending_task *pending_task = gzochid_pending_task_new (task);
  
  g_mutex_lock (task_queue->mutex);
  g_queue_insert_sorted 
    (task_queue->queue, pending_task, pending_task_compare, NULL);
  g_cond_signal (task_queue->cond);
  g_mutex_unlock (task_queue->mutex);

  return pending_task;
}

void gzochid_schedule_run_task 
(gzochid_task_queue *task_queue, gzochid_task *task)
{
  gzochid_pending_task *pending_task = 
    gzochid_schedule_submit_task (task_queue, task);

  g_mutex_lock (pending_task->mutex);
  while (pending_task->state == GZOCHID_PENDING_TASK_STATE_PENDING)
    g_cond_wait (pending_task->cond, pending_task->mutex);
  g_mutex_unlock (pending_task->mutex);
  free (pending_task);
}

void gzochid_schedule_execute_task (gzochid_task *task)
{
  task->worker (task->data, NULL);
}

==> gzochi-0.4.tar.gz: gzochi-0.4/gzochi-server/src/schedule.c <==
/* schedule.c: Task execution and task queue management routines for gzochid
 * Copyright (C) 2013 Julian Graham
 *
 * gzochi is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <glib.h>

#include "guile.h"
#include "schedule.h"
#include "task.h"

gzochid_task_queue *gzochid_schedule_task_queue_new (GThreadPool *pool)
{
  gzochid_task_queue *task_queue = malloc (sizeof (gzochid_task_queue));
  
  task_queue->cond = g_cond_new ();
  task_queue->mutex = g_mutex_new ();
  task_queue->queue = g_queue_new ();

  task_queue->pool = pool;
  task_queue->consumer_thread = NULL;

  return task_queue;
}

static void pending_task_executor (gpointer data, gpointer user_data)
{
  gzochid_pending_task *pending_task = (gzochid_pending_task *) data;

  g_mutex_lock (pending_task->mutex);
  pending_task->task->worker (pending_task->task->data, user_data);
  pending_task->state = GZOCHID_PENDING_TASK_STATE_COMPLETED;
  g_cond_broadcast (pending_task->cond);
  g_mutex_unlock (pending_task->mutex);
}

gpointer gzochid_schedule_task_executor (gpointer data)
{
  gzochid_task_queue *task_queue = (gzochid_task_queue *) data;

  g_mutex_lock (task_queue->mutex);
  while (TRUE)
    {
      if (g_queue_is_empty (task_queue->queue))
	g_cond_wait (task_queue->cond, task_queue->mutex);
      else 
	{
	  struct timeval current_time;
	  gzochid_pending_task *pending_task = 
	    (gzochid_pending_task *) g_queue_peek_head (task_queue->queue);

	  gettimeofday (&current_time, NULL);

	  if (timercmp 
	      (&current_time, &pending_task->scheduled_execution_time, >))
	    {
	      pending_task = (gzochid_pending_task *) 
		g_queue_pop_head (task_queue->queue);
	      gzochid_guile_thread_pool_push 
		(task_queue->pool, pending_task_executor, pending_task, NULL);
	    }
	  else 
	    {
	      struct timeval interval;
	      gint64 until = g_get_monotonic_time ();
	      
	      timersub 
		(&pending_task->scheduled_execution_time, 
		 &current_time, 
		 &interval);

	      until += interval.tv_sec * G_TIME_SPAN_SECOND + interval.tv_usec;
	      g_cond_wait_until (task_queue->cond, task_queue->mutex, until);
	    }
	}
    }
  g_mutex_unlock (task_queue->mutex);

  return NULL;
}

void gzochid_schedule_task_queue_start (gzochid_task_queue *task_queue)
{
  if (task_queue->consumer_thread == NULL)
    task_queue->consumer_thread = g_thread_create 
      (gzochid_schedule_task_executor, task_queue, FALSE, NULL);
}

gzochid_pending_task *gzochid_pending_task_new (gzochid_task *task)
{
  gzochid_pending_task *pending_task = malloc (sizeof (gzochid_pending_task));

  pending_task->task = task;
  pending_task->state = GZOCHID_PENDING_TASK_STATE_PENDING;
  pending_task->cond = g_cond_new ();
  pending_task->mutex = g_mutex_new ();

  pending_task->scheduled_execution_time = task->target_execution_time;
  
  return pending_task;
}

gint pending_task_compare (gconstpointer a, gconstpointer b, gpointer user_data)
{
  gzochid_pending_task *pending_task_a = (gzochid_pending_task *) a;
  gzochid_pending_task *pending_task_b = (gzochid_pending_task *) b;

  if (pending_task_a->scheduled_execution_time.tv_sec <
      pending_task_b->scheduled_execution_time.tv_sec)
    return -1;
  else if (pending_task_a->scheduled_execution_time.tv_sec 
	   > pending_task_b->scheduled_execution_time.tv_sec)
    return 1;
  else return pending_task_a->scheduled_execution_time.tv_usec 
	 - pending_task_b->scheduled_execution_time.tv_usec;
}

static void task_chain_worker (gpointer data, gpointer user_data)
{
  GList *tasks = (GList *) data;
  GList *task_ptr = tasks;

  while (task_ptr != NULL)
    {
      gzochid_schedule_execute_task ((gzochid_task *) task_ptr->data);
      task_ptr = task_ptr->next;
    }

  g_list_free (tasks);
}

gzochid_pending_task *gzochid_schedule_submit_task_chain
(gzochid_task_queue *task_queue, GList *tasks)
{
  GList *tasks_copy = g_list_copy (tasks);
  gzochid_task *task = malloc (sizeof (gzochid_task));

  assert (g_list_length (tasks) > 0);

  task->worker = task_chain_worker;
  task->data = tasks_copy;
  task->target_execution_time = 
    ((gzochid_task *) tasks_copy->data)->target_execution_time;

  return gzochid_schedule_submit_task (task_queue, task);
}

gzochid_pending_task *gzochid_schedule_submit_task
(gzochid_task_queue *task_queue, gzochid_task *task)
{
  gzochid_pending_task *pending_task = gzochid_pending_task_new (task);
  
  g_mutex_lock (task_queue->mutex);
  g_queue_insert_sorted 
    (task_queue->queue, pending_task, pending_task_compare, NULL);
  g_cond_signal (task_queue->cond);
  g_mutex_unlock (task_queue->mutex);

  return pending_task;
}

void gzochid_schedule_run_task 
(gzochid_task_queue *task_queue, gzochid_task *task)
{
  gzochid_pending_task *pending_task = 
    gzochid_schedule_submit_task (task_queue, task);

  g_mutex_lock (pending_task->mutex);
  while (pending_task->state == GZOCHID_PENDING_TASK_STATE_PENDING)
    g_cond_wait (pending_task->cond, pending_task->mutex);
  g_mutex_unlock (pending_task->mutex);
  free (pending_task);
}

void gzochid_schedule_execute_task (gzochid_task *task)
{
  task->worker (task->data, NULL);
}
